import os
import re
import pandas
import json
import argparse
import click
import matplotlib
matplotlib.use('Agg')

from itertools import islice
from collections import namedtuple



def get_result_id(f):
    assert len(f) == 47
    return f[7:-4]


def collect_data(f, id_):
    k0 = None
    c = None
    times = []
    with open(f, 'rt') as fh:
        for l in fh:
            m = re.match(r'^\s*k0 = (\d+)\s*$', l)
            if m is not None:
                n_k0 = int(m.group(1))
                if k0 is not None:
                    assert n_k0 == k0
                else:
                    k0 = n_k0
                continue

            m = re.match(r'^\s*c = (\d+)\s*$', l)
            if m is not None:
                n_c = int(m.group(1))
                if c is not None:
                    assert n_c == c
                else:
                    c = n_c
                continue

            m = re.match(r'^\s*time: (\d+.?\d*) seconds\s*$', l)
            if m is not None:
                times.append(float(m.group(1)))
    pfile = f.replace('result_', 'problem_')
    assert os.path.exists(pfile)

    n = None
    m = None
    with open(pfile, 'rt') as fh:
        problem = json.load(fh)
        mat = problem['matrix']
        n = len(mat)
        m = sum((0 if x == '-inf' else 1) for x in mat[0])

    assert len(times) == 2, "Invalid file format in " + f
    return {'id': id_, 'n': n, 'm': m, 'k0': k0, 'c': c,
            'matrix_time':times[0], 'smtdl_time':times[1]}


def get_result_files(bench_dir):
    for (dirpath, dirnames, _) in os.walk(bench_dir):
        for d in dirnames:
            for (sdirpath, _, filenames) in os.walk(os.path.join(dirpath, d)):
                for f in filenames:
                    if f.startswith('result'):
                        yield os.path.join(sdirpath, f), get_result_id(f)


def get_dataframe(dinfo, ignore_cache=False):
    if ignore_cache or not os.path.exists(dinfo.res_file):
        all_data = []
        for f, id_ in get_result_files(dinfo.bench_dir):
            data = collect_data(f, id_)
            all_data.append(data)

        df = pandas.DataFrame(all_data, columns=['id', 'n', 'm', 'k0', 'c',
                                                 'matrix_time', 'smtdl_time'])
        df.to_csv(dinfo.res_file, index=False)
        return df
    else:
        df = pandas.read_csv(dinfo.res_file)
        return df

def plot_3d(df, dinfo, show=False):
    import matplotlib.tri as mtri
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D

    x = []
    y = []
    z = {'mat':[], 'smt':[]}
    for (k0,c), data in df.groupby(['k0', 'c']):
        x.append(k0)
        y.append(c)
        z['mat'].append(data['matrix_time'].mean())
        z['smt'].append(data['smtdl_time'].mean())

    triang = mtri.Triangulation(x, y)

    for technique in ['smt', 'mat']:
        fig = plt.figure()
        ax = fig.add_subplot(1,1,1, projection='3d')

        ax.plot_trisurf(triang, z[technique], cmap='jet')
        ax.scatter(x,y,z[technique], marker='.', s=10, c="black", alpha=0.5)
        ax.view_init(elev=60, azim=-45)

        ax.set_xlabel('k0')
        ax.set_ylabel('c')
        ax.set_zlabel('runtime')
        if show:
            plt.show()
        fname = os.path.join(dinfo.plot_dir, '3d_view_%s.pdf' % technique)
        plt.savefig(fname)
        plt.close('all')


def plot_k0plusc(df, dinfo, show=False, splitby='n'):
    import matplotlib.pyplot as plt

    out_dir = os.path.join(dinfo.plot_dir, 'k0plusc')
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)

    df['k0pc'] = df['k0'] + df['c']

    for n, mdata in df.groupby(splitby):
        x = []
        mat_y = []
        smt_y = []

        for s, data in mdata.groupby(['k0pc']):
            x.append(s)
            mat_y.append(data['matrix_time'].mean())
            smt_y.append(data['smtdl_time'].mean())

        fig = plt.figure(figsize=(4.5, 3.5))
        ax = fig.add_subplot(1,1,1)

        ax.plot(x, mat_y, '--', c='blue',label='Algorithm 2')
        #ax.scatter(x, mat_y, marker='.', s=10, c='black', alpha=0.5)

        ax.plot(x, smt_y, '-', c='blue',label='Algorithm 4')
        #ax.scatter(x,smt_y, marker='.', s=10, c='black', alpha=0.5)

        locs = ax.get_yticks()
        v = [(locs[0]-locs[1])/2 for i in range(len(x))]
        ax.scatter(x,v,marker='.', s=8, c='black', alpha=0.5)

        ax.set_xlabel(r'k0+c',size='large')
        ax.set_ylabel(r'runtime(s)',size='large')

        plt.legend(loc=2)
        plt.title('Experiments with %s = %s' % (splitby, n), size='large', weight='bold')

        if show:
            plt.show()
        fname = os.path.join(out_dir, 'k0plusc_%s=%s.pdf' % (splitby, str(n).zfill(3)))
        plt.savefig(fname,bbox_inches='tight', pad_inches = 0)
        plt.close('all')

def plot_k0plusc_combine(df, dinfo, show=False):
    import matplotlib.pyplot as plt

    out_dir = os.path.join(dinfo.plot_dir, 'k0plusc')
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)

    df['k0pc'] = df['k0'] + df['c']
    splitby = ['n','m']
    SMT = []
    MAT = []
    X = []
    df.sort_values(by=['n','m','k0pc'], inplace=True,ascending=[True, True,False])
    df = df.reset_index(drop=True)
    N = sorted(list(set(df['n'])))
    NM = df[['n','m']].drop_duplicates().values.tolist()
    M = []
    num = []
    cl = ['blue','red','green','brown','yellow','magenta'] #maximum six different colors
    for i in range(len(N)):
        M_i = [NM[j][1] for j in range(len(NM)) if NM[j][0]==N[i] ]
        M.append(M_i)
        num.append(len(M_i))
    for n, mdata in df.groupby(splitby):
        x = []
        mat_y = []
        smt_y = []

        for s, data in mdata.groupby(['k0pc']):
            x.append(s)
            mat_y.append(data['matrix_time'].mean())
            smt_y.append(data['smtdl_time'].mean())

        X.append(x)
        MAT.append(mat_y)
        SMT.append(smt_y)

    it1 = iter(X)
    it2 = iter(MAT)
    it3 = iter(SMT)
    X_split = [list(islice(it1 , 0, idx)) for idx in num]
    MAT_split = [list(islice(it2 , 0, idx)) for idx in num]
    SMT_split = [list(islice(it3 , 0, idx)) for idx in num]

    print(N)
    print(M)
    print(X_split)
    for i in range(len(X_split)):
        fig = plt.figure(figsize=(4.5, 3.5))
        ax = fig.add_subplot(1,1,1)
        for j in range(len(X_split[i])):
            print((i,j))
            #ax.scatter(X_split[i][j],MAT_split[i][j], marker='.', s=10, c='black', alpha=0.6,zorder=1)
            ax.plot(X_split[i][j], MAT_split[i][j], '--',c=cl[j], label='Alg. 2, (n,m) = (%s, %s) ' % (N[i], M[i][j]))

            #ax.scatter(X_split[i][j],SMT_split[i][j], marker='.', s=10, c='black', alpha=0.6,zorder=1)
            ax.plot(X_split[i][j], SMT_split[i][j], '-',c=cl[j], label='Alg. 4, (n,m) = (%s, %s) ' % (N[i], M[i][j]))

        ax.set_xlabel(r'k0+c',size='large')
        ax.set_ylabel('runtime (s)',size='large')

        locs = ax.get_yticks()
        for j in range(len(X_split[i])):
            v = [(locs[0]-locs[1])/2 for k in range(len(X_split[i][j]))]
            ax.scatter(X_split[i][j],v,marker='.', s=8, c='black', alpha=0.5)

        plt.legend(loc=2,labelspacing=.2)
        plt.title('Experiments with n = %s' % N[i], size='large', weight='bold')

        if show:
            plt.show()
        fname = os.path.join(out_dir, 'k0plusc_combined_%s.pdf' % N[i])
        plt.savefig(fname,bbox_inches='tight', pad_inches = 0)
        plt.close('all')


def plot_k0plusc_n_m(df, dinfo, show=False):
    import matplotlib.pyplot as plt

    out_dir = os.path.join(dinfo.plot_dir, 'k0plusc')
    if not os.path.exists(out_dir):
        os.mkdir(out_dir)

    df['k0pc'] = df['k0'] + df['c']
    df.sort_values(by=['n','m','k0pc'], inplace=True,ascending=[True, True,False])
    df = df.reset_index(drop=True)
    splitby = ['n','m']
    cl = ['blue','red','green','brown','yellow','magenta'] #maximum six different colors
    N = sorted(list(set(df['n'])))
    NM = df[['n','m']].drop_duplicates().values.tolist()
    M = []
    for i in range(len(N)):
        M_i = [NM[j][1] for j in range(len(NM)) if NM[j][0]==N[i] ]
        M.append(M_i)

    for n, mdata in df.groupby(splitby):
        x = []
        mat_y = []
        smt_y = []
        check = mdata['matrix_time'] < mdata['smtdl_time']
        idx = next(iter(check.index[check]), 'not exist')
        cross = mdata.loc[idx-1,'k0pc']

        for s, data in mdata.groupby(['k0pc']):
            x.append(s)
            mat_y.append(data['matrix_time'].mean())
            smt_y.append(data['smtdl_time'].mean())

        fig = plt.figure(figsize=(4.5, 3.5))
        ax = fig.add_subplot(1,1,1)
        ax.text(0.32, 0.95,r'cross-over point = %s' %cross,horizontalalignment='left',verticalalignment='center',transform = ax.transAxes,size='large')

        idx1 = N.index(n[0])
        idx2 = M[idx1].index(n[1])
        #ax.scatter(x, mat_y, marker='.', s=10, c='black', alpha=0.6,zorder=1)
        ax.plot(x, mat_y, '--', c=cl[idx2],label='Alg. 2')

        #ax.scatter(x,smt_y, marker='.', s=10, c='black', alpha=0.6,zorder=1)
        ax.plot(x, smt_y, '-', c=cl[idx2], label='Alg. 4')

        locs = ax.get_yticks()
        v = [(locs[0]-locs[1])/2 for i in range(len(x))]
        ax.scatter(x,v,marker='.', s=8, c='black', alpha=0.5)

        ax.set_xlabel(r'k0+c',size='large')
        ax.set_ylabel('runtime (s)',size='large')

        plt.legend(loc=2,labelspacing=.2)
        plt.title('Experiments with (n,m) = (%s, %s)' % (n), size='large', weight='bold')

        if show:
            plt.show()
        fname = os.path.join(out_dir, 'k0plusc_%s.pdf' % str(n).zfill(3))
        plt.savefig(fname,bbox_inches='tight', pad_inches = 0)
        plt.close('all')


def plot_k0plusc_overall(df, dinfo, show=False):
    import matplotlib.pyplot as plt

    df['k0pc'] = df['k0'] + df['c']

    x = []
    mat_y = []
    smt_y = []

    for s, data in df.groupby(['k0pc']):
        x.append(s)
        mat_y.append(data['matrix_time'].mean())
        smt_y.append(data['smtdl_time'].mean())

    fig = plt.figure(figsize=(4.5, 3.5))
    ax = fig.add_subplot(1,1,1)

    #ax.scatter(x, mat_y, marker='.', s=10, c='black', alpha=0.5)
    ax.plot(x, mat_y, '--', c='blue',label='Algorithm 2')

    #ax.scatter(x,smt_y, marker='.', s=10, c='black', alpha=0.5)
    ax.plot(x, smt_y, '-', c='blue',label='Algorithm 4')

    locs = ax.get_yticks()
    v = [(locs[0]-locs[1])/2 for i in range(len(x))]
    ax.scatter(x,v,marker='.', s=8, c='black', alpha=0.5)

    ax.set_xlabel('k0+c',size='large')
    ax.set_ylabel('runtime (s)',size='large')

    plt.legend(loc=2)
    plt.title('Overall experiments', size='large', weight='bold')

    if show:
        plt.show()
    fname = os.path.join(dinfo.plot_dir, 'k0plusc.pdf')
    plt.savefig(fname,bbox_inches='tight', pad_inches = 0)
    plt.close('all')


DirInfo = namedtuple('DirInfo', ['bench_dir', 'res_file', 'plot_dir'])

@click.command()
@click.option('-i', '--ignore-cache', is_flag=True,
              help='Force rebuild the cache CSV')
@click.option('-s', '--show', is_flag=True,
              help='Show the plots in a GUI while they are generated')
@click.option('-e', '--enable-syifaul-plots', 'enable_extra_plots', is_flag=True,
              help='Force rebuild the cache CSV')
@click.option('-E', '--enable-3d-plot', 'enable_3d_plots', is_flag=True,
              help='Force rebuild the cache CSV')

@click.argument('BENCH_DIR', type=click.Path(exists=True))
def main(bench_dir, ignore_cache=False, show=False, enable_extra_plots=False,
         enable_3d_plots=False):
    '''
    Generate the results of the benchmark contained in BENCH_DIR
    '''
    bench_dir = os.path.realpath(bench_dir)
    res_file = os.path.join(bench_dir, 'results.csv')
    plot_dir = os.path.join(bench_dir, 'plots')

    if not os.path.exists(plot_dir):
        os.mkdir(plot_dir)

    dinfo = DirInfo(bench_dir, res_file, plot_dir)

    df = get_dataframe(dinfo, ignore_cache=ignore_cache)

    plot_k0plusc(df, dinfo, splitby='n')
    plot_k0plusc_overall(df, dinfo, show=show)

    if enable_3d_plots:
        plot_3d(df, dinfo, show=show)

    if enable_extra_plots:
        plot_k0plusc_combine(df, dinfo, show=show)
        plot_k0plusc_n_m(df, dinfo, show=show)


if __name__ == '__main__':
    main()
