import sys
if sys.version_info >= (3, 3):
    from collections.abc import Iterable
else:
    from collections import Iterable

from pysmt.shortcuts import And, Or, Plus, Not, Equals, GE, LE, Ite, Implies, Bool
from pysmt.shortcuts import Real, Minus, Symbol, REAL, Solver, GT, LT

from pympl import matrix
from pympl import simulation
from pympl import utils




def Max(*args):
    """Returns the encoding of the maximum expression within args"""
    if len(args) == 1 and isinstance(args[0], Iterable):
        args = args[0]
    exprs = tuple(args)

    assert len(exprs) > 0
    if len(exprs) == 1:
        return exprs[0]
    elif len(exprs) == 2:
        a, b = exprs
        return Ite(LE(a, b), b, a)
    else:
        h = len(exprs) // 2
        cond = And(Or(GE(o, x) for o in exprs[0:h]) for x in exprs[h:])
        return Ite(cond, Max(exprs[0:h]), Max(exprs[h:]))

def trans_list(A, v):
    res = []
    for r in A:
        c = min(e for e in r if e is not None)
        t = Max((Plus(Real(e-c), v[i]) if e != c else v[i])
                for i,e in enumerate(r) if e is not None)
        res.append((c, t))
    return res

def trans_noite(A, v, v1):
    res = []
    for i,r in enumerate(A):
        terms = [Plus(Real(e), v[j]) for j,e in enumerate(r) if e is not None]
        res.append(Or(LE(v1[i], t) for t in terms))  #I change Equals to LE
        for t in terms:
            res.append(GE(v1[i], t))
    return And(res)

def trans_noite_sym(A, v, v1):
    res = []
    for i,r in enumerate(A):
        terms = [(ii, Plus(e, v[j])) for j,(ii,e) in enumerate(r)]
        res.append(Or(And(Not(ii), LE(v1[i] ,t)) for ii,t in terms))  #I change Equals to LE
        for ii, t in terms:
            res.append(Implies(Not(ii), GE(v1[i], t)))
    return And(res)


def get_loop_formula(v1, v2):
    last_diff = None
    for x,y in zip(v1, v2):
        diff = Minus(x, y)
        if last_diff is not None:
            yield Equals(last_diff, diff)
        last_diff = diff

def get_loop_formula_c(v1, v2):
    min_c_diff = None
    min_idx = None
    for i, ((c1, x), (c2, y)) in enumerate(zip(v1, v2)):
        c_diff = c1 - c2
        if min_c_diff is None or c_diff < min_c_diff:
            min_c_diff = c_diff
            min_idx = i
    min_diff = Minus(v1[min_idx][1], v2[min_idx][1])
    for i, ((c1, x), (c2, y)) in enumerate(zip(v1, v2)):
        if i != min_diff:
            c_diff = c1 - c2 - min_c_diff
            diff = Minus(x, y)
            if c_diff != 0:
                yield Equals(Plus(diff, Real(c_diff)), min_diff)
            else:
                yield Equals(diff, min_diff)

def get_negated_loop_formula_c(v1, v2):
    for eq in get_loop_formula_c(v1, v2):
        yield GT(eq.arg(0), eq.arg(1))
        yield LT(eq.arg(0), eq.arg(1))

def smt_period_single_vars(A, solver=None):
    k0, c = 0, 1
    vrs = [Symbol('x%d' % i, REAL) for i in range(len(A))]
    cache = {}
    while True:
        #print(" Trying with k0=%d and c=%d" % (k0, c))
        A1 = matrix.maxplus_pwr(A, k0, cache)
        A2 = matrix.maxplus_pwr(A, c+k0, cache)

        v1 = trans_list(A1, vrs)
        v2 = trans_list(A2, vrs)

        n_loop = Or(Not(x) for x in get_loop_formula_c(v1, v2))

        with Solver(name=solver) as s:
            s.add_assertion(n_loop)
            if not s.solve():
                return k0, c
            else:
                k0, c = refine_from_model(A, s, vrs, k0, c)

def smt_period_single_vars_distrib(A, solver=None):
    k0, c = 0, 1
    vrs = [Symbol('x%d' % i, REAL) for i in range(len(A))]
    cache = {}
    while True:
        print("Trying with k0=%d and c=%d" % (k0, c))
        A1 = matrix.maxplus_pwr(A, k0, cache)
        A2 = matrix.maxplus_pwr(A, c+k0, cache)

        v1 = trans_list(A1, vrs)
        v2 = trans_list(A2, vrs)

        tbc = False
        for x in get_negated_loop_formula_c(v1, v2):
            with Solver(name=solver) as s:
                s.add_assertion(x)
                print('.', end='')
                if s.solve():
                    k0, c = refine_from_model(A, s, vrs, k0, c)
                    tbc = True
                    break
        if not tbc:
            return k0, c


def refine_from_model(A, s, v0, k0, c):
    s0 = []
    for v in v0:
        s0.append(s.get_py_value(v))

    try:
        b_k0 = simulation.back_simulate(A, s0)
    except TypeError:
        b_k0 = 0

    n_k0, n_c, _ = simulation.simulate(A, s0)
    assert n_k0 > k0 or n_c > c, "No progress! There is a BUG!"
    k0, c = max(k0, n_k0+b_k0), utils.lcm(c, n_c)
    return k0, c


def smt_period_basic(A, solver=None):
    k0, c = 0, 1
    v0 = [Symbol('x%d' % i, REAL) for i in range(len(A))]
    v1 = [Symbol('y%d' % i, REAL) for i in range(len(A))]
    v2 = [Symbol('z%d' % i, REAL) for i in range(len(A))]
    cache = {}
    while True:
        print(" Trying with k0=%d and c=%d" % (k0, c))
        A1 = matrix.maxplus_pwr(A, k0, cache)
        A2 = matrix.maxplus_pwr(A, c, cache)

        n_loop = Or(Not(x) for x in get_loop_formula(v1, v2))

        with Solver(name=solver) as s:
            s.add_assertion(trans_noite(A1, v0, v1))
            s.add_assertion(trans_noite(A2, v1, v2))
            s.add_assertion(n_loop)
            if not s.solve():
                return k0, c
            else:
                k0, c = refine_from_model(A, s, v0, k0, c)

def assign_mat(M, V):
    for i in range(len(M)):
        for j in range(len(M)):
            if V[i][j] is not None:
                yield Equals(M[i][j][1], Real(V[i][j]))
                yield Not(M[i][j][0])
            else:
                yield Equals(M[i][j][1], Real(0))
                yield M[i][j][0]

def smt_period_sym_matrix(A, solver=None):
    k0, c = 0, 1
    v0 = [Symbol('x%d' % i, REAL) for i in range(len(A))]
    v1 = [Symbol('y%d' % i, REAL) for i in range(len(A))]
    v2 = [Symbol('z%d' % i, REAL) for i in range(len(A))]

    B = [[(Symbol('inf_b%d_%d' % (i,j)), Symbol('b%d_%d' % (i,j), REAL))
          for j in range(len(A))] for i in range(len(A))]
    C = [[(Symbol('inf_c%d_%d' % (i,j)), Symbol('c%d_%d' % (i,j), REAL))
          for j in range(len(A))] for i in range(len(A))]

    cache = {}
    with Solver(name=solver) as s:
        s.add_assertion(trans_noite_sym(B, v0, v1))
        s.add_assertion(trans_noite_sym(C, v1, v2))
        n_loop = Or(Not(x) for x in get_loop_formula(v1, v2))
        s.add_assertion(n_loop)
        while True:
            print(" Trying with k0=%d and c=%d" % (k0, c))
            A1 = matrix.maxplus_pwr(A, k0, cache)
            A2 = matrix.maxplus_pwr(A, c, cache)

            assumptions = list(assign_mat(B, A1))
            assumptions += list(assign_mat(C, A2))

            if not s.solve(assumptions=assumptions):
                return k0, c
            else:
                k0, c = refine_from_model(A, s, v0, k0, c)

def strict_ineq_func(A,B):
    m = len(A)
    n = len(A[0])
    expr = []
    var = [Symbol('x%d' % i, REAL) for i in range(len(A[0]))]
    k = 0
    stop = False
    while (k<m and (not stop)):
        A_k = [A[k][i] if A[k][i] is not None and (B[k][i] is None or A[k][i]>=B[k][i]) else None for i in range(n)]
        B_k = [B[k][i] if B[k][i] is not None and (A[k][i] is None or B[k][i]> A[k][i]) else None for i in range(n)]
        idx1 = [i for i in range(n) if A_k[i] is not None]
        idx2 = [i for i in range(n) if B_k[i] is not None]
        expr_k = None
        if len(idx1)==0:
            stop = True
            expr = []
        elif len(idx1)>0 and len(idx2)>0:
            expr_k = And(Or(GE(Minus(var[idx1[i]],var[idx2[j]]),Real(B_k[idx2[j]]-A_k[idx1[i]])) for i in range(len(idx1))) for j in range(len(idx2)))
            expr.append(expr_k)
        k += 1
    if len(expr)==0 and (not stop):
        return Bool(True)
    elif len(expr)==0 and stop:
        return Bool(False)
    else:
        return(And(x for x in expr))

def equality_func(A,B):
    F1 = strict_ineq_func(A,B)
    F2 = strict_ineq_func(B,A)
    if F1==Bool(False) or F1==Bool(False):
        return Bool(False)
    elif F1==Bool(True):
        return F2
    elif F2==Bool(True):
        return F1
    else:
        return And(F1,F2)

def smt_period_rdl(A, init=False,solver=None):
    k0, c = 0, 1
    var = [Symbol('x%d' % i, REAL) for i in range(len(A))]
    cache = {}
    A1 = matrix.maxplus_pwr(A, k0, cache)
    A2 = matrix.maxplus_pwr(A, k0+c, cache)
    w = matrix.cycle_mean(A)
    A1 = matrix.maxplus_scaling(A1,w)
    if init:
        I = GE(Minus(var[0],var[1]),Real(0))
        for i in range(len(A)-2):
            I = And(I,GE(Minus(var[i+1],var[i+2]),Real(0)))
    else:
        I = Bool(True)
    with Solver(name=solver) as s:
        while True:
            #print(" Trying with k0=%d and c=%d" % (k0, c))
            F = Not(equality_func(A1,A2))
            if not s.solve(assumptions=[I,F]):
                return k0, c
            else:
                s0 = []
                for v in var:
                    s0.append(s.get_py_value(v))
                s1 = []
                for i in range(len(A1)):
                    s1.append([max(A1[i][j] + s0[j] for j in range(len(s0)) if A1[i][j] is not None)])
                #n_k0, n_c, _ = simulation.simulate(A, s1)
                n_k0, n_c = matrix.maxplus_period(A,s1)
                assert n_k0 > 0 or n_c > c, "No progress! There is a BUG!"
                n_c = utils.lcm(c, n_c)
                if (n_k0>0):
                    A1 = matrix.maxplus_times(A1,matrix.maxplus_pwr(A,n_k0,cache))
                    A2 = matrix.maxplus_times(A2,matrix.maxplus_pwr(A,n_k0,cache))
                    k0 = n_k0 + k0
                if (n_c>c):
                    A2 = matrix.maxplus_times(A2,matrix.maxplus_pwr(A,n_c-c,cache))
                    A1 = matrix.maxplus_scaling(A1,w*(n_c-c))
                    c = n_c
